#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <limits.h>

#include "instance.h"
#include "utilities.h"
#include "timer.h"
#include "ls.h"

static double uups;
#define ABS(x) ((uups=(x))>0?(uups):(-uups))

#define TIMING

int main (int argc, char **argv) 

{

  /* remark: tours are represented as a permutation of the city
indices 0 to n-1 (i.e., n cities); additionally, at the position n,
the index of the first city is repeated. This eases teh computation of
tour length etc. This is also the reason why the malloc below
allocates a vector of length ncities + 1. */

  int i, j, try;
  int *tour;     /* current tour for ILS */
  int *shadow;   /* modified tour after double-bridge move */
  int *random;   /* a random vector */
  int length;    /* length of tour */
  char temp_buffer[100];

  setbuf(stdout,NULL);
  
  if (argc < 2) {
    fprintf (stderr, "Usage:  test filename\n");
  }

  seed = (long int) time(NULL); /* initialize random number generator */
  
  seed = 500;
 
#ifdef TIMING
  start_timers(); /* starts time measurement */
#endif
  
  readInstance(argv[1]);
  
  sprintf(temp_buffer,"best");
  report = fopen(temp_buffer, "w");
  sprintf(temp_buffer,"cmp");
  comp_report = fopen(temp_buffer, "w");
  fprintf(comp_report,"begin problem %s\n",argv[1]);


#ifdef TIMING
/*    printf ("Time to read instance = %4.2f\n\n", ABS(elapsed_time( VIRTUAL ))); */
#endif
  
  distMat = computeDistances();
/*    printDistances(); */
#ifdef TIMING
/*    printf ("Cumulative time up to compute distances = %4.2f\n\n", elapsed_time( VIRTUAL )); */
#endif

  nn_ls = 40;

  nnMat = computeNNLists();

#ifdef TIMING
  start_timers(); /* starts time measurement */
#endif

  tour = malloc((ncities + 1 ) * sizeof(int));

  for ( try = 1; try <= 25; try ++ ) {
#ifdef TIMING
/*        start_timers(); */
#endif
      fprintf(comp_report,"begin try %d\n",try);
      random =   construct_nn_tour();
      for ( j = 0 ; j < ncities ; j++ ) {
	  tour[j] = random[j];
      }
      tour[ncities] = tour[0];
      dlb = calloc(ncities, sizeof(int));
#ifdef TIMING
/*        start_timers(); */
#endif
      three_opt_first(tour);
/*        length = computeLength( tour ); */
/*        printf("\nbest %d\t time %f\t iteration %d\n",length,fabs(elapsed_time( VIRTUAL )),0); */

#ifdef TIMING
/*        printf ("Cumulative time up to compute NN lists = %4.2f\n\n", elapsed_time( VIRTUAL )); */
#endif
      
      for ( i = 0 ; i < 10 * ncities ; i++ ) {

	  shadow = doublebridge_window ( tour );
	  three_opt_first( shadow );
	  accept( tour, &length, shadow, i );
	  free (shadow);
      }
      
      free ( random );
      free( dlb );
      fprintf(report,"best %d\t time %f\t iteration %d\t total.time %f\n",length, time_best_found, iteration_best_found,fabs(elapsed_time( VIRTUAL )));
      fprintf(comp_report,"end try %d\n",try);
      length = LONG_MAX;
  }
  free ( tour );

#ifdef TIMING
/*    printf ("Cumulative time to run ILS = %4.2f\n\n", elapsed_time( VIRTUAL )); */
#endif

/*    printf("END, SUCCESSFUL!!\n\n"); */
  fprintf(comp_report,"end problem %s\n",argv[1]);

  return 0;
}






