#include <stdio.h>
#include <stdlib.h>

#include "utilities.h"
#include "instance.h"

long int seed;

void fatal (char *s) {
  fprintf (stderr, "%s\n", s);
  exit(1);
}


void **createMatrix (int i, int j, int s) {

  int k;
  char **result = (char **) (malloc(i*sizeof(result) + i*j*s));
  char *p = (char *)result + (i*sizeof(result));

  for (k=0; k<i; ++k) {
    result[k] = p;
    p += j*sizeof(s);
  }

  return (void **)result;
}

double ran01( long *idum ) {
/* 
      FUNCTION:      returns a pseudo-random number
      INPUT:         a pointer to the seed variable 
      OUTPUT:        a pseudo-random number uniformly distributed in [0,1]
      (SIDE)EFFECTS: changes the value of seed
*/
  long k;
  double ans;

  k =(*idum)/IQ;
  *idum = IA * (*idum - k * IQ) - IR * k;
  if (*idum < 0 ) *idum += IM;
  ans = AM * (*idum);
  return ans;
}


long int random_number( long *idum )
/*    
      FUNCTION:       generate an integer random number
      INPUT:          pointer to variable containing random number seed
      OUTPUT:         integer random number uniformly distributed in {0,2147483647}
      (SIDE)EFFECTS:  random number seed is modified (important, has to be done!)
      ORIGIN:         numerical recipes in C
*/
{
  long k;

  k =(*idum)/IQ;
  *idum = IA * (*idum - k * IQ) - IR * k;
  if (*idum < 0 ) *idum += IM;
  return *idum;
}


int * generate_random_vector()
     /* Generates a random vector, quick and dirty */
{
   int     *random_vector;
   int     i, help, node, tot_assigned = 0;
   double  rnd;

   random_vector = malloc(ncities * sizeof(int));  

   for ( i = 0 ; i < ncities; i++) 
     random_vector[i] = i;

   for ( i = 0 ; i < ncities ; i++ ) {
     /* find (randomly) an index for a free unit */ 
     rnd  = ran01 ( &seed );
     node = (long int) (rnd  * (ncities - tot_assigned)); 
     help = random_vector[i];
     random_vector[i] = random_vector[i+node];
     random_vector[i+node] = help;
     tot_assigned++;
   }

   return random_vector;

}

void swap(int v[], int v2[], int i, int j)
{
  long int tmp;

  tmp = v[i];
  v[i] = v[j];
  v[j] = tmp;
  tmp = v2[i];
  v2[i] = v2[j];
  v2[j] = tmp;
  
}

void sort(int v[], int v2[], int left, int right)
{
  long int k, last;

  if (left >= right) 
    return;
  swap(v, v2, left, (left + right)/2);
  last = left;
  for (k=left+1; k <= right; k++)
    if (v[k] < v[left])
      swap(v, v2, ++last, k);
  swap(v, v2, left, last);
  sort(v, v2, left, last);
  sort(v, v2, last+1, right);
}
